@@ -394,6 +394,7 @@ class TestRegressions3 (unittest.TestCase):
         self.mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
         self.mf.run_script(os.path.join(root, 'script.py'))
 
+    @unittest.skipUnless(not hasattr(sys, 'real_prefix'), "Test doesn't work in virtualenv")
     def testRegr1(self):
         node = self.mf.findNode('mypkg.distutils')
         self.assertIsInstance(node, modulegraph.Package)
@@ -454,7 +455,103 @@ class TestRegression5 (unittest.TestCase):
     def testRegr1(self):
         node = self.mf.findNode('distutils')
         self.assertIsInstance(node, modulegraph.Package)
-        self.assertIn('distutils/__init__', node.filename)
+        self.assertIn(os.path.join('distutils', '__init__'), node.filename)
+
+class TestDeeplyNested (unittest.TestCase):
+    def setUp(self):
+        root = os.path.join(
+                os.path.dirname(os.path.abspath(__file__)),
+                'testpkg-regr6')
+        self.mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
+        self.mf.run_script(os.path.join(root, 'script.py'))
+
+    def testRegr(self):
+        node = self.mf.findNode('os')
+        self.assertIsNot(node, None)
+
+
+class TestRelativeReferenceToToplevel (unittest.TestCase):
+    if not hasattr(unittest.TestCase, 'assertIsInstance'):
+        def assertIsInstance(self, value, types):
+            if not isinstance(value, types):
+                self.fail("%r is not an instance of %r"%(value, types))
+
+    def test_relative_import_too_far(self):
+        # pkg.mod tries to import "..sys" (outside of the package...)
+        root = os.path.join(
+                os.path.dirname(os.path.abspath(__file__)),
+                'testpkg-regr7')
+        mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
+        mf.run_script(os.path.join(root, 'script.py'))
+
+        m = mf.findNode('')
+        self.assertIs(m, None)
+
+        m = mf.findNode('pkg.mod')
+        self.assertIsInstance(m, modulegraph.SourceModule)
+
+        imported = list(mf.get_edges(m)[0])
+        self.assertEqual(len(imported), 5)
+
+        im = imported[0]
+        self.assertIsInstance(im, modulegraph.InvalidRelativeImport)
+        self.assertEqual(im.relative_path, '..')
+        self.assertEqual(im.from_name, 'sys')
+        self.assertEqual(im.identifier, '..sys')
+
+        im1 = imported[1]
+        im2 = imported[2]
+        if im1.identifier == '...xml':
+            # Order of modules imported in a single 'from .. import a, b' list
+            # is unspecified, ensure a fixed order for this test.
+            im2, im1 = im1, im2
+
+        self.assertIsInstance(im1, modulegraph.InvalidRelativeImport)
+        self.assertEqual(im1.relative_path, '...')
+        self.assertEqual(im1.from_name, 'os')
+        self.assertEqual(im1.identifier, '...os')
+
+        im = imported[2]
+        self.assertIsInstance(im2, modulegraph.InvalidRelativeImport)
+        self.assertEqual(im2.relative_path, '...')
+        self.assertEqual(im2.from_name, 'xml')
+        self.assertEqual(im2.identifier, '...xml')
+
+        im = imported[3]
+        self.assertIsInstance(im, modulegraph.InvalidRelativeImport)
+        self.assertEqual(im.relative_path, '..foo')
+        self.assertEqual(im.from_name, 'bar')
+        self.assertEqual(im.identifier, '..foo.bar')
+
+        im = imported[4]
+        self.assertIs(im, mf.findNode('pkg'))
+
+class TestInvalidAsyncFunction (unittest.TestCase):
+    if not hasattr(unittest.TestCase, 'assertIsInstance'):
+        def assertIsInstance(self, value, types):
+            if not isinstance(value, types):
+                self.fail("%r is not an instance of %r"%(value, types))
+
+    @unittest.skipUnless(sys.version_info[:2] == (3,5), "Requires python 3.5")
+    def test_invalid_async_function(self):
+        # In python 3.5 the following function is invalid:
+        #
+        #    async def foo():
+        #       yield 1
+        #
+        # This is a syntax error that's reported when compiling the AST
+        # to bytecode, which caused an error in modulegraph.
+        #
+        # In python 3.6 this is valid code (and in earlier versions async
+        # versions didn't exist)
+        root = os.path.join(
+                os.path.dirname(os.path.abspath(__file__)),
+                'testpkg-regr8')
+        mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
+        mf.run_script(os.path.join(root, 'script.py'))
+
+        n = mf.findNode('mod')
+        self.assertIsInstance(n, modulegraph.InvalidSourceModule)
 
 if __name__ == "__main__":
     unittest.main()
